package org.eclipse.palamedes.gdl.connection;
import java.io.IOException;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

public class PlayerServer extends NanoHTTPD {

	/* Default settings */
	protected static int 	 DEFAULT_PORT = 4001;
	protected static boolean DEFAULT_MODE = false;
	
    /** Stores the player of this server */
    protected Player  player;
    protected int     port;
    protected boolean slaveMode = DEFAULT_MODE;

    /* Internal used strings */
    final static String STR_SERVER_START = "Started a PlayerServer - ";
    final static String STR_ERROR1       = "Message is empty!";
    final static String STR_ERROR2       = "Unknown command ";
    final static String STR_ERROR3       = "Unknown message state ";

    public PlayerServer(Player player, Map<String,String> options) 
    throws IOException {
        super(getPort(options));
        this.player = player;
        this.setOptions(options);
        System.out.println(STR_SERVER_START + "SlaveMode[" + slaveMode+  "]");
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.connection.NanoHTTPD#serve(java.lang.String, java.lang.String, java.util.Properties, java.util.Properties, java.lang.String)
     */
    @Override
    public Response serve( String uri, String method, Properties header, Properties parms, String data ) {
        try {
            String response_string = "";

            if ( data != null ) {
                System.out.println( "Command: " + data );

                // parse message
                Message message = new Message(data);

                // switch to correct function
                switch ( message.type ) {
                case Message.START:
                    player.commandStart( message );
                    response_string = "READY";
                    break;
                case Message.PLAY:
                    response_string = player.commandPlay( message );
                    if (slaveMode)
                        response_string += player.getSlaveExtention();
                    break;
                case Message.STOP:
                    player.commandStop( message );
                    response_string = "DONE";
                    break;
                case Message.EMPTY:
                    throw( new IllegalArgumentException(STR_ERROR1) );
                case Message.ERROR:
                    throw( new IllegalArgumentException(STR_ERROR2 + data) );
                default:
                    throw( new IllegalArgumentException(STR_ERROR3 + data) );
                }
            } else {
                System.out.println( "Empty message received!" );
            }

            System.out.println( "Response:" + response_string );

            Response response = new Response( Response.HTTP_OK, "text/acl", response_string );
            response.addHeader( "Content-Length", "" + response_string.length() );
            return response;

        // error handling
        } catch ( IllegalArgumentException ex ) {
            ex.printStackTrace();
            return new Response( Response.HTTP_BADREQUEST, "text/acl", "NIL" );
        }
    }


    /**
     * wait for server thread to exit
     *
     */
    public void waitForExit(){
        try {
            server_thread.join();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    /**
     * This method is a hook to setup any relevant commandline option within the
     * player server or player. 
     * @param options
     */
    protected void setOptions(Map<String, String> options) {
    	/* Slave Option */
    	String slaveValue = options.get("slave");
    	if (slaveValue != null)
    		if ( "true".equalsIgnoreCase(slaveValue) ) {
    			slaveMode = true;
    		} else if ("false".equalsIgnoreCase(slaveValue)) {
    			slaveMode = false;
    		} else {
    			System.err.println("Slave argument is wrong. It should be true or false. ");
    			System.exit(1);
    		}
    }
    
    /**
     * This method creates and returns the option map.
     */
    public static Map<String, String> getOptions(String args[]) {
    	TreeMap<String, String> argumentMap = 
			  new TreeMap<String,String>(String.CASE_INSENSITIVE_ORDER);

    	for (String argument: args) 
    		if  (argument.startsWith("--")) {
    			String[] argumentSplit = argument.split("=");
    			argumentMap.put(argumentSplit[0].substring(2), argumentSplit[1]);
    		}
    	return argumentMap;
    }
    

    /**
     * This method extracts the port number out of the argument map.
     * @param argumentMap
     */
    protected static int getPort(Map<String, String> argumentMap) {
    	int    port 	 = PlayerServer.DEFAULT_PORT;
    	String portValue = argumentMap.get("port");
    	if (portValue != null) {
    		try {
    			port = Integer.parseInt(portValue);
    		} catch (NumberFormatException ex) {
    			System.err.println("Port argument is wrong. It should be between 1 to 65535.");
    			System.exit(1);
    		}
    	}
    	return port;
    }
    
    /**
     * starts the game player and waits for messages from the game master <br>
     * Command line options: [port] slave
     */
    public static void main(String[] args){

    	/* create and start player server */
    	try {    		
            new PlayerServer(new Player(), getOptions(args)).waitForExit();
        } catch (IOException ex) {
            ex.printStackTrace();
            System.exit(-1);
        }
    }
    
}
