package org.eclipse.palamedes.gdl.core.knowledgebase.analyser;

import java.util.ArrayList;
import java.util.HashSet;

import org.eclipse.palamedes.gdl.core.ast.Init;
import org.eclipse.palamedes.kif.core.ast.ASTMatcher;
import org.eclipse.palamedes.kif.core.ast.ASTNode;
import org.eclipse.palamedes.kif.core.ast.KIFFunTerm;
import org.eclipse.palamedes.kif.core.ast.KIFImplication;
import org.eclipse.palamedes.kif.core.ast.KIFIndVar;
import org.eclipse.palamedes.kif.core.ast.KIFRelSent;
import org.eclipse.palamedes.kif.core.ast.KIFSent;
import org.eclipse.palamedes.kif.core.ast.KIFSeq;
import org.eclipse.palamedes.kif.core.ast.KIFTerm;
import org.eclipse.palamedes.kif.core.ast.visitor.GroundChecker;
import org.eclipse.palamedes.kif.core.knowledgebase.IKBAnalyser;
import org.eclipse.palamedes.kif.core.knowledgebase.IKBFilter;
import org.eclipse.palamedes.kif.core.knowledgebase.KBEntry;
import org.eclipse.palamedes.kif.core.knowledgebase.KnowledgeBase;


public class AUnaryFunctions implements IKBAnalyser {

    static class AUFFilter implements IKBFilter {

        public boolean isMatching(KBEntry entry) {
            return KBEntry.V_FUN.equals( entry.get(KBEntry.P_TYPE)  ) &&
                   "1".equals          ( entry.get(KBEntry.P_ARITY) );
        }

    }

    private KnowledgeBase currentKB;

    public void analyse(KnowledgeBase kb) {

        currentKB = kb;
        for (KBEntry entry : kb.filter( new AUFFilter() ) ) {
        	try {
        		findToggles(entry);
	            findCounter(entry);
	            findControlToggle(entry);
	            findStepCounter(entry);
        	} catch(Exception ex) {
        		System.err.println("Exception in game analysation");
        		ex.printStackTrace();
        	}
        }

    }


    /**
     * This method returns whether the given function is a toggle one or not.
     *
     * So it breaks down to:
     *      A toggle is:
     *          - a unary function
     *          - only implications with a next relation exist (rules)
     *          - rules are ground
     *
     * Complexity: O(n) with n = |implicationHeadNodes|
     */
    private void findToggles(KBEntry info) {
        GroundChecker checker = new GroundChecker();

        // iterate over headNodes and determine the domain of the toggle
        KIFRelSent head;
        KIFRelSent body;
        KIFFunTerm funcHead;
        KIFFunTerm funcBody;
        HashSet<String> domain0 = new HashSet<String>();

        for ( KIFImplication node : info.getImplicationHead() ) {
            if ( !checker.isGround(node) ||
                 node.getPremise().size() != 1 )
                return;

            if (node.getConsequence().getNodeType() != ASTNode.RELSENT)
            	return;
            head = (KIFRelSent) node.getConsequence();

            KIFSent firstPremise = node.getPremise().get(0);
            if ( firstPremise.getNodeType() != ASTNode.RELSENT)
            	return;
            
            body = (KIFRelSent) firstPremise;

            if ( !"next".equalsIgnoreCase(head.getName()) ||
                 !"true".equalsIgnoreCase(body.getName()) )
                return;

            KIFTerm firstHeadArg = head.getTerms().get(0);
            KIFTerm firstBodyArg = body.getTerms().get(0);
            if ( firstHeadArg.getNodeType()!=ASTNode.FUNTERM ||
            	 firstBodyArg.getNodeType()!=ASTNode.FUNTERM)
            	return;
            
            funcHead = (KIFFunTerm) firstHeadArg;
            funcBody = (KIFFunTerm) firstBodyArg;
            domain0.add( funcHead.getTerms().get(0).getName() );
            domain0.add( funcBody.getTerms().get(0).getName() );
        }

        // we survived ? so it is a toggle
        info.put("isToggle", 		true);
        info.put("domain_distinct", domain0);
    }

    /**
     * This method returns whether the given function is a counter one or not.
     *
     * A counter rule is defined as following (infix notation):
     * <code>
     *      next(<func> (?1) ) <= true( <func>(?2) ) ^ <domain>( ?2, ?1 )
     * </code>
     *
     * So it breaks down to:
     *      A counter ...
     *      - a unary function
     *      - has only 1 rule
     *      - operates on a domain with arity 2
     *      - the body of counter rule consists of ...
     *          - one true predicate with the counter function
     *          - one strict ordered domain
     */
    private void findCounter(KBEntry info) {

        // we are looking for a function that has only one rule
        if ( info.getImplicationHead().size() != 1 )
            return;

        // inner inspection
        KIFImplication  implication = info.getImplicationHead().get(0);
        if(implication.getConsequence().getNodeType()!=ASTNode.RELSENT)
        	return;
        KIFRelSent 	    head	    = (KIFRelSent) implication.getConsequence();
        KIFSeq<KIFSent> body	    = implication.getPremise();

        // should be a next rule
        if ( !"next".equalsIgnoreCase( head.getName() ) )
            return;

        // test head and body size
        if ( head.getTerms().size() != 1 || body.size() != 2 )
            return;

        // get variables
        KIFSent p1 = body.get(0);
        KIFSent p2 = body.get(1);

        // swap variables -> we want true in p1 and domain in p2
        if ( "true".equalsIgnoreCase( p2.getName() ) ) {
            p2 = p1;
            p1 = null;
        }

        // check domain
        KBEntry info2 = currentKB.get(p2.getName());
        if ( !info2.has("hasStrictOrder") )
            return;

        Object domainDistinct = info2.get("domain_distinct");
        info.put("isCounter", 		true);
        info.put("domain_distinct", domainDistinct);
        info.put("successor",       info2.getName());


        //
        // getting the direction
        //

        // head variable position 0 points to ...
        KIFTerm t = head.getTerms().get(0);
        if (t.getNodeType()!=ASTNode.FUNTERM)
        	return;
        KIFFunTerm funTerm = (KIFFunTerm) t;
        t = funTerm.getTerms().get(0);
        if (t.getNodeType()!=ASTNode.INDVAR)
        	return;
        KIFIndVar  varPos0 = (KIFIndVar) t;
        String     varNamePos0 = varPos0.getName();

        // ... body variable position ...
        if (p2.getNodeType()!=ASTNode.RELSENT)
        	return;
        KIFRelSent relation = (KIFRelSent) p2;
        String     varBody0 = relation.getTerms().get(0).getName();

        // ... 0 ?
        if ( varBody0.equalsIgnoreCase(varNamePos0) ) {
            info.put("countStart", info2.get("domain_1_end"));
            info.put("countEnd",   info2.get("domain_0_end"));
        } else {
            info.put("countStart", info2.get("domain_0_end"));
            info.put("countEnd",   info2.get("domain_1_end"));
        }

    }

    /**
     * This method looks for control toggles.
     * That means: a control toggle is a toggle that operates on a role domain.
     *
     * @param info
     */
    @SuppressWarnings("unchecked")
    private void findControlToggle(KBEntry info) {

        // only toggles are relevant
        if ( !info.has("isToggle") )
            return;

        // check domains
        Object a = info.get("domain_distinct");
        Object b = currentKB.get("role").get("domain_distinct");

        HashSet<String> toggleDomain = (HashSet<String>) a;
        HashSet<String> roleDomain   = (HashSet<String>) b;

        // 1:1 mapping between toggle domain and role domain
        if ( roleDomain.containsAll(toggleDomain) &&
             toggleDomain.containsAll(roleDomain) )
            info.put("isControlToggle", true);
    }

    /**
     * This method looks for step counter.
     * That means: a step counter is started by an init relation and is stopped
     * 			   by a terminal rule.
     *
     * @param info
     */
    private void findStepCounter(KBEntry info) {

        // only counter are relevant
        if ( !info.has("isCounter") )
            return;

        // check that we appear exactly once within init
        ArrayList<KIFSent> inits = currentKB.get("init").getTopLevelNodes();
        String  name;
        boolean found = false;
        for (ASTNode node : inits) {
            name = ((Init) node).getArgument().getName();
            if ( info.getName().equalsIgnoreCase(name) ) {

                // should be only once
                if (found)
                    return;

                found = true;
            }
        }

        // at least one time it should appear within the init fluents
        if (!found)
            return;

        ArrayList<KIFImplication> terminals = currentKB
                                                    .get("terminal")
                                                        .getImplicationHead();
        found = false;
        for (KIFImplication rule : terminals) {
            for (ASTNode node : rule.getPremise()) {

                if ( !(node instanceof KIFRelSent) )
                    continue;

                name = ((KIFRelSent) node).getTerms().get(0).getName();
                if ( info.getName().equalsIgnoreCase(name) ) {

                    // should be only once
                    if (found)
                        return;

                    found = true;
                }
            }
        }

        // at least one time it should appear within a terminal rule
        if (!found)
            return;

        // we survived ? so it is a step counter
        info.put("isStepCounter", true);

    }
}
