package org.eclipse.palamedes.gdl.core.model;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;

import org.eclipse.palamedes.gdl.core.model.utils.AbstractFactory;

/**
 * The GameFactory class is responsible for loading and creating games from all
 * kind of sources with the reasoner of choice.
 * 
 * It offers two methods:
 * <ul>
 * 	<li>loadGame(<ReasonerType>, <filename>): which loads a game from a file </li>
 *  <li>createGame(<ReasonerType>, <gdlString>): which creates a game from a string</li>
 * </ul>
 */
public final class GameFactory extends AbstractFactory {

	/** Singleton instance holder class for thread-safe lazy instantiation. */
    private static class InstanceHolder { 
    	private static GameFactory instance = new GameFactory();
    }
	
    private static final String pkg = "org.eclipse.palamedes.gdl.core.resolver"; 
    
    public static final String JAVAPROVER 	= "JavaProver";
    public static final String JOCULAR    	= "Jocular";
    public static final String PROLOGPROVER = "PrologProver";

    /**
     * Private constructor to avoid instantiation outside this class.
     */
    private GameFactory() {    	
        addDescription( JAVAPROVER,
        				pkg + ".javaprover.GameJP",
				  		"JavaProver Adapter" );
        addDescription( JOCULAR,
						pkg + ".jocular.GameJocular",
		  				"Jocular Adapter" );
        addDescription( PROLOGPROVER,
						pkg + ".prologprover.GameProlog",
		  				"PrologProver Adapter" );
    }

    /**
     * Returns the instance of the strategy factory.
     */
    public static GameFactory getInstance() {
        return InstanceHolder.instance;
    }
    
    public IGame createGame(String reasoner, String gdl) {
        try {
        	IGame game = (IGame) createInstance(getDescriptionByName(reasoner));
        	game.init(gdl);
        	return game;
        } catch (Exception e) {        	
        	System.err.println("Could not create the game: " + e.getMessage());
        	e.printStackTrace();
        }
        return null;
    }

    
    public IGame loadGame(String reasoner, String filename) {
        return createGame(reasoner, getGDL(filename));
    }

    
    /**
     * This method loads the gdl string from a file location.
     * @param location
     * @return
     */
    private String getGDL(String location) {
        if (location == null || location.length() == 0)
            return null;

        String       encoding = "";
        StringBuffer buffer   = new StringBuffer();
        try {
            FileInputStream   fStream  = new FileInputStream(location);
            InputStreamReader iStream  = new InputStreamReader(fStream);

            encoding = iStream.getEncoding();
            iStream  = new InputStreamReader(fStream, encoding);

            Reader in = new BufferedReader(iStream);
            int ch;
            while ( (ch = in.read()) > -1 ) {
                buffer.append((char)ch);
            }
            in.close();
            return buffer.toString();
        }
        catch (UnsupportedEncodingException e) {
            System.err.println("GameFactory.getGDL()");
            System.err.println("Unsupported Encoding: " + encoding);
        }
        catch (FileNotFoundException e) {
            System.err.println("GameFactory.getGDL()");
            System.err.println("FileNotFound: " + location);
        }
        catch (IOException e) {
            System.err.println("GameFactory.getGDL()");
            System.err.println("IOException while reading file: " + location);
        }
        return "";
    }
    
}
