package org.eclipse.palamedes.gdl.core.model.utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.palamedes.gdl.core.model.IFluent;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;


public abstract class AbstractState<T> implements IGameState {

    protected final T                          nativeState;
    protected List<IFluent>                    fluents;
    protected Map<String, List<IFluent>>       fluentMap;

    IMove[][]                                  legalMovesArray;
    boolean                                    terminal;
    int[]                                      goals;


    public AbstractState(T state) {
        nativeState = state;
        terminal    = false;
    }

    
    abstract public void initFluentList();

    
    /**
     * Returns and creates the Fluent Name Mapping.
     */
    private final Map<String, List<IFluent>> getFluentMap() {

        if (fluentMap == null) {
            fluentMap = new HashMap<String, List<IFluent>>();

            // construct filter map
            for ( IFluent fluent : getFluents() ) {
                String fluentName = fluent.getName();
                if ( !fluentMap.containsKey(fluentName) )
                    fluentMap.put(fluentName, new ArrayList<IFluent>());
                fluentMap.get(fluentName).add(fluent);
            }
        }
        return fluentMap;
    }
    
    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getCombinedLegalMoves()
     */
    public List<IMove[]> getCombinedLegalMoves() {
        return legalMovesArray != null ? new JoinedList(legalMovesArray) : null;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getFluentByName(java.lang.String)
     */
    public List<IFluent> getFluentByName(String name) {
        return getFluentMap().get(name);
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getFluents()
     */
    public List<IFluent> getFluents() {
        if (fluents == null) {
            fluents = new ArrayList<IFluent>();
            initFluentList();
        }
        return fluents;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getFluentNames()
     */
    public List<String> getFluentNames() {
        return new ArrayList<String>(getFluentMap().keySet());
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getGoalValues()
     */
    public int[] getGoalValues() {
        return (goals != null) ? goals.clone() : null;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#isTerminal()
     */
    public boolean isTerminal() {
        return terminal;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#setGoalValue(java.lang.String, int)
     */
    public int getGoalValue(int roleIndex) {
        return (goals != null) ? goals[roleIndex] : -1;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#setGoalValue(java.lang.String, int)
     */
    public void setGoalValues(int[] goalValues) {
        goals = goalValues.clone();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#setTerminal()
     */
    public void setTerminal() {
        terminal = true;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getNativeState()
     */
    public T getNativeState() {
        return nativeState;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#getLegalMoves()
     */
    public IMove[][] getLegalMoves() {
        return legalMovesArray;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGameState#setLegalMoves(org.eclipse.palamedes.gdl.core.model.IMove[][])
     */
    public void setLegalMoves(IMove[][] legalMovesArray) {
        this.legalMovesArray = legalMovesArray;
    }

    @Override
    public int hashCode() {
        return nativeState.hashCode();
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj) {
        return (obj == this) || 
               ( (obj instanceof AbstractState) && 
               ((AbstractState) obj).nativeState.equals(this.nativeState) );
    }

}
