package org.eclipse.palamedes.gdl.core.model.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Stack;

import org.eclipse.palamedes.gdl.core.ast.GDLAST;
import org.eclipse.palamedes.gdl.core.knowledgebase.GDLKnowledgeBase;
import org.eclipse.palamedes.gdl.core.knowledgebase.analyser.AEnvironment;
import org.eclipse.palamedes.gdl.core.knowledgebase.analyser.AUnaryFunctions;
import org.eclipse.palamedes.gdl.core.knowledgebase.filter.StepCounterFilter;
import org.eclipse.palamedes.gdl.core.model.IGame;
import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IGameTree;
import org.eclipse.palamedes.gdl.core.model.IMove;
import org.eclipse.palamedes.gdl.core.model.IReasoner;
import org.eclipse.palamedes.gdl.core.model.IStatistic;
import org.eclipse.palamedes.kif.core.ast.KIFForm;
import org.eclipse.palamedes.kif.core.knowledgebase.KBEntry;
import org.eclipse.palamedes.kif.core.knowledgebase.analyser.ABinaryFacts;
import org.eclipse.palamedes.kif.core.knowledgebase.analyser.ADomainInformation;

public abstract class Game implements IGame {

    /** A random number generator. */
    protected static final Random random = new Random();

    // statistic values - method calls
    int cLegalMovesCall  = 0;
    int cNextStateCall   = 0;
    int cNextNodeCall    = 0;

    // calculation time
    long cLegalMovesTime = 0;
    long cNextStateTime  = 0;

    // content count
    int cLegalMoves      = 0;

    /** Game class private stuff **/
    private List<String> roleList;

    /** Game class protected stuff **/
    protected IGameTree        tree;
    protected IReasoner        reasoner;

    // generated content
    protected GDLKnowledgeBase kb;
    protected String           sourceGDL;
    protected String[]         roleNames;
    protected IGameNode		   initNode;

    public Game() {    	
    }
    
    public void init(IReasoner initReasoner, IGameTree initTree) {
        this.tree     = initTree;
        this.reasoner = initReasoner;
    }

    /**
     * This method returns the game tree.
     */
    public IGameTree getTree() {
        return tree;
    }

    public final String getSourceGDL() {
        return sourceGDL;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getRoleNames()
     */
    public final String[] getRoleNames() {
        return roleNames.clone();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getInitialNode()
     */
    public final IGameNode getInitialNode() {
    	return tree.getRootNode();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getRoleCount()
     */
    public int getRoleCount() {
        return roleNames.length;
    }

    public int getRoleIndex(String role) {
        if (roleList == null)
            roleList = Arrays.asList(roleNames);

        return roleList.indexOf(role);
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getCombinedMoves(org.eclipse.palamedes.gdl.core.model.IGameNode)
     */
    public List<IMove[]> getCombinedMoves(IGameNode currentNode)
    throws InterruptedException {

        IGameState state = getState(currentNode);
        
        // if already resolved get the legal moves of the currentNode
        if ( state.getLegalMoves() != null )
            return state.getCombinedLegalMoves();

        resolveMoves(state);
        
        // and now the state should be able to return the combined legal moves
        return state.getCombinedLegalMoves();
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getLegalMoves(org.eclipse.palamedes.gdl.core.model.IGameNode)
     */
    public IMove[][] getLegalMoves(IGameNode currentNode)
			throws InterruptedException {

        IGameState state = getState(currentNode);
        
        // if already resolved get the legal moves of the currentNode
        if ( state.getLegalMoves() != null )
            return state.getLegalMoves();

        resolveMoves(state);
        
        // and now the state should be able to return the combined legal moves
        return state.getLegalMoves();
	}

	private void resolveMoves(IGameState state)
    throws InterruptedException {
        
        // generate all moves for all players and set them back to the state
        IMove[][] moves = new IMove[roleNames.length][];
        for (int i = 0, j = roleNames.length; i < j; i++) {
            cLegalMovesCall++;
            long start       = System.currentTimeMillis();
            moves[i]         = reasoner.getLegalMoves(roleNames[i], state);
            cLegalMovesTime += (System.currentTimeMillis() - start);
            cLegalMoves     += moves[i].length;
        }

        // store them back to the node
        state.setLegalMoves(moves);
    }
    
    
    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getNextNode(org.eclipse.palamedes.gdl.core.model.IGameNode, org.eclipse.palamedes.gdl.core.model.IMove[])
     */
    public IGameNode getNextNode(IGameNode currentNode, IMove[] moves)
    throws InterruptedException {
        cNextNodeCall++;

        // fetch the child node
        IGameNode nextNode = tree.getNextNode(currentNode, moves);

        // if already resolved get the next node of the currentNode
        if ( nextNode != null && nextNode.getState() != null )
            return nextNode;

        // otherwise create the next node
        IGameState currentState = getState(currentNode);

        cNextStateCall++;
        long start           = System.currentTimeMillis();
        IGameState nextState = reasoner.getNextState(currentState, moves);
        cNextStateTime      += System.currentTimeMillis() - start;

        // if the state is terminal then determine if it has a goal value
        if ( nextState.isTerminal() ) {

            // announce the terminal state to the tree
            tree.setTerminal(nextState);

            // resolve the goal value
            resolveGoalValues(nextState);
        }

        // if we already had a node but the state was empty then we set the
        // state and return the node.
        if (nextNode != null)
            nextNode.setState(nextState);
        else
            nextNode = tree.addChild(currentNode, nextState, moves);

        // and return it to the caller
        return nextNode;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getRandomMove(org.eclipse.palamedes.gdl.core.model.IGameNode)
     */
    public IMove[] getRandomMove(IGameNode currentNode)
    throws InterruptedException {

        IGameState state = getState(currentNode);
        
        // generate moves if moves are missing
        if ( state.getLegalMoves() == null)
            resolveMoves(state);
        
        // build a random move ...
        IMove[][] moves = state.getLegalMoves();
        IMove[]   move  = new IMove[roleNames.length];
        for (int i = 0, j = roleNames.length; i < j; i++)
            move[i] = moves[i][random.nextInt(moves[i].length)];
        
        // ... and return it
        return move;
    }

    public int[] getGoalValues(IGameNode currentNode)
			throws InterruptedException {

        IGameState state = getState(currentNode);
        
        // generate goal values if they are missing
        if ( state.getGoalValues() == null)
        	resolveGoalValues(state);
        
        return state.getGoalValues();
	}

	public boolean isTerminal(IGameNode currentNode)
			throws InterruptedException {

		return currentNode.isTerminal();
	}

	private void resolveGoalValues(IGameState state)
			throws InterruptedException {
		
        // resolve the goal value
        int[] goalValues = new int[roleNames.length];
        int   i          = 0;
        for (String roleName : getRoleNames()) {
            goalValues[i] = reasoner.getGoalValue(roleName, state);
            i++;
        }
        state.setGoalValues(goalValues);
	}
	
	public IGameState getState(IGameNode currentNode)
			throws InterruptedException {

		IGameState state=currentNode.getState();
		
		if(state==null){
			regenerateNode(currentNode);
			state=currentNode.getState();
		}

		return state;
	}

	/*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#regenerateNode(org.eclipse.palamedes.gdl.core.model.IGameNode)
     */
    public void regenerateNode(IGameNode node)
    throws InterruptedException {

        long start = System.currentTimeMillis();

        IGameNode parent = node.getParent();

        if (parent == null)
            return;

        Stack<IGameNode> nodeStack = new Stack<IGameNode>();

        // fill stack
        IGameNode currentNode = node;
        while ( currentNode            != null &&
                currentNode.getState() == null &&  
                parent                 != null ) {

            nodeStack.push(currentNode);
            currentNode = parent;
            parent      = parent.getParent();
        }

        // regenerate nodes from top to bottom
        while ( !nodeStack.isEmpty() ) {

            // get top node
            currentNode = nodeStack.pop();
            getNextNode( currentNode.getParent(), currentNode.getMoves() );
        }

        // get the possible moves for the requested node
        System.err.println( "regeneration duration (ms) " +
                            (System.currentTimeMillis() - start) );
    }

    public IStatistic getStatistic() {
        StatisticInfo info = new StatisticInfo();
        info.setProperty(IStatistic.S_MOVE_COUNT, "" + cLegalMoves);
        info.setProperty(IStatistic.S_MOVE_CALL,  "" + cLegalMovesCall);
        info.setProperty(IStatistic.S_MOVE_TIME,  "" + cLegalMovesTime);
        info.setProperty(IStatistic.S_STATE_CALL, "" + cNextStateCall);
        info.setProperty(IStatistic.S_STATE_TIME, "" + cNextStateTime);
        info.setProperty(IStatistic.S_NODE_CALL,  "" + cNextNodeCall);
        return info;
    }


    public GDLKnowledgeBase getKB() {
        if (kb != null)
            return kb;

        String gdl = getSourceGDL();
        kb = new GDLKnowledgeBase();
        kb.setDescription( (KIFForm) new GDLAST().parseString(gdl) );

        try {
            new ADomainInformation().analyse(kb);
            new ABinaryFacts().analyse(kb);
            new AUnaryFunctions().analyse(kb);
            new AEnvironment().analyse(kb);
        } catch (Exception exp) {
            System.err.println("Error while knowledge base analysis.");
            exp.printStackTrace();
        }

        return kb;
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getMaxSteps()
     */
    public int getMaxSteps() {
        List<KBEntry> counter = getKB().filter(new StepCounterFilter());
        for (KBEntry entry : counter) {

            List<?> end    = (List<?>) entry.get("countEnd");
            KBEntry succ   = kb.get((String) entry.get("successor"));
            List<?> domain = (List<?>) succ.get("domain_1");

            List<String> converted = new ArrayList<String>();
            for (Object item : domain)
                converted.add( item.toString() );

            return converted.indexOf( end.get(0) ) + 1;
        }
        return -1;
    }


    public static String showGameNodePath(IGameNode node) {
        StringBuilder sb = new StringBuilder();

        Stack<IGameNode> nodeStack = new Stack<IGameNode>();
        nodeStack.push(node);

        IGameNode parent = node.getParent();
        while ( parent != null ) {
            nodeStack.push(parent);
            parent = parent.getParent();
        }

        while ( !nodeStack.isEmpty() )
            sb.append(nodeStack.pop()).append("\n");

        return sb.toString();
    }


    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getReasoner()
     */
    public IReasoner getReasoner() {
        return reasoner;
    }
    
    /*
     * (non-Javadoc)
     * @see org.eclipse.palamedes.gdl.core.model.IGame#getPerftValue(int, org.eclipse.palamedes.gdl.core.model.IGameNode)
     */
    public int getPerftValue(int depth, IGameNode startNode) 
    throws InterruptedException {

    	// there is no successor node for negative steps;
    	if (depth < 0)
    		return -1;
    	
    	switch (depth) {
    	
    	// perft(0) should be 0 since there are no successor nodes
    	case 0:
    		return 0;
    		
    	// perft(1) corresponds to the moves available in the startNode
    	case 1:
    		return getCombinedMoves(startNode).size();

    	}
    	
    	// perft(depth): here the real work starts
    	int perft = 0;
    	
    	// initialise node and move stack
    	Stack<IGameNode> 		 nodeStack = new Stack<IGameNode>();
    	Stack<Iterator<IMove[]>> moveStack = new Stack<Iterator<IMove[]>>();
    	
    	// do not forget the serialisation of joined list!
    	List<IMove[]> moves = getCombinedMoves(startNode);
    	nodeStack.push(startNode);
    	moveStack.push(moves.iterator());
    	
    	// depth-first search to keep memory footprint small
    	while ( !moveStack.isEmpty() ) {

    		// look at the top moves
    		Iterator<IMove[]> currentMoves = moveStack.peek();
    	
    		// if level reached just count the moves and pop them
    		if (nodeStack.size() == depth) {
    			moves = getCombinedMoves(nodeStack.peek());
	    		perft += moves.size();
    			
    			moveStack.pop();
				nodeStack.pop();
    		    		
    		} else 
    			
    			// we do not have moves left -> fall back to previous level
    			if (!currentMoves.hasNext()) {
    				moveStack.pop();
    				nodeStack.pop();

    			} else {

    				// expand move
    		    	nodeStack.push( getNextNode( nodeStack.peek(), currentMoves.next() ) );
    		    	moveStack.push( getCombinedMoves(nodeStack.peek()).iterator() );
    		}
    		
    	}
    	return perft;
    }
    
}
