package org.eclipse.palamedes.gdl.core.simulation;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.palamedes.gdl.core.model.IGame;
import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;


public class Simulator {

    // simulation info - static
    private final IGame           game;
    private final Match[]         matchs;

    // simulation info - dynamic
    private SimulationSettings   settings;
    private int                  step;
    private IGameNode            currentNode;
    private IGameState           currentState;
    private final IMove[]        currentMoves;
    private StepInfo             currentStepInfo;
    private final List<StepInfo> stepHistory  = new ArrayList<StepInfo>();


    public Simulator(IGame newGame) {
        super();

        // generate the matches
        game         = newGame;
        matchs       = new Match[game.getRoleCount()];
        currentMoves = new IMove[game.getRoleCount()];

        // Initialize dynamic structures
        reset();
    }

    public void run() {
        while (!isTerminal())
            step();
    }

    public void setSettings(SimulationSettings newSettings) {
        settings = newSettings;
    }

    /**
     * Creates a default random setting in the case of missing external
     * settings. Its a kind of fall-back setting.
     */
    private void createRandomSettings() {
        int playerCount = game.getRoleCount();

        // init all arrays
        IStrategy[] strategies = new IStrategy[playerCount];
        int[]       startTimes = new int[playerCount];
        int[]       playTimes  = new int[playerCount];

        for (int i = 0; i < playerCount; i++) {
            strategies[i] = StrategyFactory.getInstance().createStrategy(0);
            startTimes[i] = -1;
            playTimes[i]  = -1;
        }

        settings = new SimulationSettings(strategies, startTimes, playTimes);
    }

    /**
     * Creates the matches and initializes the runtime information
     */
    public void reset() {
        // create random settings if none are available
        if (settings == null)
            createRandomSettings();

        // create match objects
        for (int i = 0, j = game.getRoleCount(); i < j; i++)
            matchs[i] = new Match( game,
                                   settings.strategies[i],
                                   game.getRoleNames()[i],
                                   settings.startTimes[i],
                                   settings.playTimes[i] );

        // clear history
        for (StepInfo info : stepHistory)
            info.clear();
        stepHistory.clear();

        // set step and currentNode to start
        step         = 0;
        currentNode  = game.getTree().getRootNode();
        currentState = currentNode.getState();
    }

    public void step() {

        hookBeforeMoveGeneration();
        generateMoves();
        hookAfterMoveGeneration();

        // simulation bookkeeping
        generateStepInfo();
        stepHistory.add(currentStepInfo);
        step++;

        hookBeforeStateGeneration();
        generateNextState();
        hookAfterStateGeneration();

        if ( isTerminal() ) {

            // empty next moves
            hookBeforeMoveGeneration();
            for (int i = 0; i < game.getRoleCount(); i++)
                currentMoves[i] = null;
            hookAfterMoveGeneration();

            // add the last state to history
            generateStepInfo();
            stepHistory.add(currentStepInfo);
            step++;

            hookIsTerminal();
        }
    }

    /**
     * This method iterates over all player and calls match.getMove().
     */
    private void generateMoves() {
        for (int i = 0, j = game.getRoleCount(); i < j; i++ )
            currentMoves[i] = matchs[i].getMove();
    }

    /**
     * This method iterates over all matches, applies the moves and stores the
     * next state.
     */
    private void generateNextState() {
        for (Match match : matchs)
            match.makeTurn(currentMoves);
        currentNode  = matchs[0].getCurrentNode();
        currentState = currentNode.getState();
    }

    private void generateStepInfo() {

        // preparing data
        List<IMove> copyMoves = new ArrayList<IMove>();
        for (IMove move : currentMoves)
            copyMoves.add(move);

        // start creating info
        currentStepInfo = new StepInfo();
        currentStepInfo.put(StepInfo.STEP,  step                       );
        currentStepInfo.put(StepInfo.MOVES, copyMoves                  );
        currentStepInfo.put(StepInfo.NODE,  matchs[0].getCurrentNode() );
        currentStepInfo.put(StepInfo.STATE, currentState               );
        currentStepInfo.put(StepInfo.STATS, game.getStatistic()        );
    }

    public List<StepInfo>   getStepInfos()      { return stepHistory;   }

    // getter for internal state
    public IGame       getGame()        { return game;                        }
    public Match[]    getMatchs()       { return matchs.clone();              }
    public int        getStep()         { return step;                        }
    public IMove[]    getCurrentMoves() { return currentMoves.clone();        }
    public IGameState getCurrentState() { return currentState;                }
    public boolean    isTerminal()      { return currentState.isTerminal();   }

    // empty hook methods
    public void hookBeforeMoveGeneration() {}
    public void hookAfterMoveGeneration()  {}
    public void hookBeforeStateGeneration(){}
    public void hookAfterStateGeneration() {}
    public void hookIsTerminal()           {}


    public void setCurrentState(IGameState currentState) {
        this.currentState = currentState;
    }


    public void setCurrentNode(IGameNode currentNode) {
        for (Match match : matchs)
            match.currentNode = currentNode;
        currentState = currentNode.getState();
    }
}
