package org.eclipse.palamedes.gdl.core.simulation.strategies;

import java.util.List;
import java.util.Stack;

import org.eclipse.palamedes.gdl.core.model.IGameNode;
import org.eclipse.palamedes.gdl.core.model.IGameState;
import org.eclipse.palamedes.gdl.core.model.IMove;
import org.eclipse.palamedes.gdl.core.simulation.Match;
import org.eclipse.palamedes.gdl.core.simulation.TimerFlag;


public class SDFS extends AbstractStrategy {

    /** contains the best path to a goal so far found */
    int[]   bestPath;
    double  bestValue;


    @Override
    public void initMatch(Match initMatch) {
        super.initMatch(initMatch);
        bestPath = new int[maxSteps];
        search(game.getTree().getRootNode());
    }

    private void search(IGameNode startNode) {

        try {

            // localize information for speed up
            TimerFlag timer = match.getTimer();

            // initializes the moves from this node
            // -> has to be known before creating the corresponding FringeFrame
            game.getCombinedMoves(startNode);

            // create fringe stack and put starting node on it
            Stack<FringeFrame> fringe = new Stack<FringeFrame>();
            fringe.push( new FringeFrame(startNode) );

            // copy bestPath to preserve old decisions
            int[] path = bestPath.clone();

            // start with the assumption that the best path is not known
            bestValue = -1;

            // run until stopped by time
            while ( !fringe.isEmpty() && bestValue < 100 ) {

                if ( timer.interrupted() )
                    return;

                // get top stack frame and state
                FringeFrame topFrame = fringe.peek();
                IGameNode   node     = topFrame.node;

                // generate next moves
                if (node.getState() == null)
                    game.regenerateNode(node);

                // find next unvisited move && set it to be visited
                int index = topFrame.movesForExpansion.nextSetBit(0);

                // if all moves are explored -> remove the node from fringe
                if (index == -1) {
                    fringe.pop();
                    continue;
                }

                topFrame.movesForExpansion.clear(index);
                path[node.getDepth()] = index;

                // generate next state
                IMove[]    moves    = game.getCombinedMoves(node).get(index);
                IGameNode  newNode  = game.getNextNode(node, moves);
                IGameState newState = newNode.getState();

                // test the new node
                if ( newState.isTerminal() ) {

                    if ( newState.getGoalValue(playerNumber) > bestValue ) {
                        bestPath  = path;
                        bestValue = newState.getGoalValue(playerNumber);
                    }
                    fringe.pop();

                    // still not terminal -> next iteration on this node
                } else {

                    // initializes the moves from this node
                    // -> has to be known before creating the corresponding
                    // FringeFrame
                    game.getCombinedMoves(newNode);
                    fringe.push( new FringeFrame(newNode) );
                }
            }
        } catch (InterruptedException e) {
//          System.out.println("Search stopped by time.");
        }
    }

    @Override
    public IMove getMove(IGameNode currentNode) {
        search(currentNode);

        try {
            // get best next move
            int           index = bestPath[currentNode.getDepth()];
            List<IMove[]> moves = match.getGame().getCombinedMoves(currentNode);

            System.out.println( ">>> SDFS Move Selection");
            System.out.println( "bestGoalValue[" + bestValue +
                                "], move[" + moves.get(index)[playerNumber] +
                                "]");
            return moves.get(index)[playerNumber];
        }
        catch (InterruptedException e) {
        }
        return null;
    }

}
